#!/usr/bin/env python

__all__ = [
    "blast",
    "cigar",
    "clustal",
    "dialign",
    "ebi",
    "fasta",
    "gcg",
    "genbank",
    "gff",
    "locuslink",
    "ncbi_taxonomy",
    "newick",
    "nexus",
    "paml",
    "paml_matrix",
    "phylip",
    "rdb",
    "record",
    "record_finder",
    "sequence",
    "table",
    "tinyseq",
    "tree",
    "tree_xml",
    "unigene",
]

__author__ = ""
__copyright__ = "Copyright 2007-2022, The Cogent Project"
__credits__ = [
    "Gavin Huttley",
    "Peter Maxwell",
    "Rob Knight",
    "Catherine Lozupone",
    "Jeremy Widmann",
    "Matthew Wakefield",
    "Sandra Smit",
    "Greg Caporaso",
    "Zongzhi Liu",
    "Micah Hamady",
    "Jason Carnes",
    "Raymond Sammut",
    "Hua Ying",
    "Andrew Butterfield",
    "Marcin Cieslik",
]
__license__ = "BSD-3"
__version__ = "2023.2.12a1"
__maintainer__ = "Gavin Huttley"
__email__ = "Gavin.Huttley@anu.edu.au"
__status__ = "Production"
