#!/usr/bin/env python
__all__ = [
    "test_best_likelihood",
    "test_bootstrap",
    "test_distance",
    "test_likelihood_function",
    "test_motifchange",
    "test_parameter_controller",
    "test_scale_rules",
    "test_simulation",
    "test_substitution_model",
    "test_coevolution",
    "test_models",
]

__author__ = ""
__copyright__ = "Copyright 2007-2022, The Cogent Project"
__credits__ = ["Rob Knight", "Peter Maxwell", "Greg Caporaso"]
__license__ = "BSD-3"
__version__ = "2023.2.12a1"
__maintainer__ = "Gavin Huttley"
__email__ = "Gavin.Huttley@anu.edu.au"
__status__ = "Production"
