#!/usr/bin/env python
__all__ = [
    "test_geometry",
    "test_matrix_logarithm",
    "test_measure",
    "test_optimisers",
    "test_stats",
    "test_util",
]

__author__ = ""
__copyright__ = "Copyright 2007-2022, The Cogent Project"
__credits__ = [
    "Rob Knight",
    "Peter Maxwell",
    "Catherine Lozupone",
    "Gavin Huttley",
    "Sandra Smit",
    "Marcin Cieslik",
    "Antonio Gonzalez Pena",
]
__license__ = "BSD-3"
__version__ = "2023.2.12a1"
__maintainer__ = "Gavin Huttley"
__email__ = "Gavin.Huttley@anu.edu.au"
__status__ = "Production"
