#!/usr/bin/env python

__all__ = ["dendrogram", "dotplot", "drawable", "letter", "logo"]

__copyright__ = "Copyright 2007-2022, The Cogent Project"
__contributors__ = [
    "Peter Maxwell",
    "Gavin Huttley",
    "Rob Knight",
    "Zongzhi Liu",
    "Matthew Wakefield",
    "Stephanie Wilson",
    "Rahul Ghangas",
    "Sheng Han Moses Koh",
]
__license__ = "BSD-3"
__version__ = "2023.2.12a1"
__status__ = "Production"
