#!/usr/bin/env python
__all__ = [
    "checkpointing",
    "deserialise",
    "dict_array",
    "misc",
    "parallel",
    "progress_display",
    "recode_alignment",
    "table",
    "transform",
    "union_dict",
    "warning",
]

__author__ = ""
__copyright__ = "Copyright 2007-2022, The Cogent Project"
__credits__ = [
    "Gavin Huttley",
    "Rob Knight",
    "Sandra Smit",
    "Peter Maxwell",
    "Amanda Birmingham",
    "Zongzhi Liu",
    "Andrew Butterfield",
    "Daniel McDonald",
    "Greg Caporaso",
    "Thomas La",
]
__license__ = "BSD-3"
__version__ = "2023.2.12a1"
__maintainer__ = "Gavin Huttley"
__email__ = "gavin.huttley@anu.edu.au"
__status__ = "Production"
