__all__ = [
    "test_dotplot",
    "test_dendrogram",
    "test_draw_integration",
    "test_logo",
    "test_shapes",
]

__author__ = "Gavin Huttley and Rahul Ghangas"
__copyright__ = "Copyright 2007-2022, The Cogent Project"
__credits__ = ["Gavin Huttley"]
__license__ = "BSD-3"
__version__ = "2023.2.12a1"
__maintainer__ = "Gavin Huttley"
__email__ = "Gavin.Huttley@anu.edu.au"
__status__ = "Alpha"
