#!/usr/bin/env python

__all__ = [
    "test_alignment",
    "test_alphabet",
    "test_annotation",
    "test_core_standalone",
    "test_features",
    "test_genetic_code",
    "test_info",
    "test_location",
    "test_maps",
    "test_moltype",
    "test_profile",
    "test_sequence",
    "test_tree",
]

__author__ = ""
__copyright__ = "Copyright 2007-2022, The Cogent Project"
__credits__ = [
    "Catherine Lozupone",
    "Peter Maxwell",
    "Rob Knight",
    "Gavin Huttley",
    "Jeremy Widmann",
    "Greg Caporaso",
    "Sandra Smit",
    "Justin Kuczynski",
]
__license__ = "BSD-3"
__version__ = "2023.2.12a1"
__maintainer__ = "Gavin Huttley"
__email__ = "Gavin.Huttley@anu.edu.au"
__status__ = "Production"
