#!/usr/bin/env python

__all__ = [
    "alignment",
    "alphabet",
    "annotation",
    "genetic_code",
    "info",
    "location",
    "moltype",
    "profile",
    "sequence",
    "tree",
]

__author__ = ""
__copyright__ = "Copyright 2007-2022, The Cogent Project"
__credits__ = [
    "Jeremy Widmann",
    "Gavin Huttley",
    "Rob Knight",
    "Sandra Smit",
    "Peter Maxwell",
    "Matthew Wakefield",
    "Greg Caporaso",
]
__license__ = "BSD-3"
__version__ = "2023.2.12a1"
__maintainer__ = "Gavin Huttley"
__email__ = "gavin.huttley@anu.edu.au"
__status__ = "Production"
