#!/usr/bin/env python

__all__ = [
    "geometry",
    "markov",
    "matrix_exponential_integration",
    "matrix_exponentiation",
    "matrix_logarithm",
    "optimisers",
    "period",
    "scipy_optimisers",
    "scipy_optimize",
    "simannealingoptimiser",
    "solve",
    "util",
]


__author__ = ""
__copyright__ = "Copyright 2007-2022, The Cogent Project"
__credits__ = [
    "Gavin Huttley",
    "Peter Maxwell",
    "Matthew Wakefield",
    "Rob Knight",
    "Edward Lang",
    "Sandra Smit",
    "Antonio Gonzalez Pena",
    "Ben Kaehler",
]
__license__ = "BSD-3"
__version__ = "2023.2.12a1"
__maintainer__ = "Gavin Huttley"
__email__ = "gavin.huttley@anu.edu.au"
__status__ = "Production"
