#!/usr/bin/env python
__all__ = ["test_bedgraph", "test_clustal", "test_fasta"]

__author__ = ""
__copyright__ = "Copyright 2007-2022, The Cogent Project"
__credits__ = [
    "Rob Knight",
    "Gavin Huttley",
    "Sandra Smit",
    "Marcin Cieslik",
    "Jeremy Widmann",
]
__license__ = "BSD-3"
__version__ = "2023.2.12a1"
__maintainer__ = "Gavin Huttley"
__email__ = "gavin.huttley@anu.edu.au"
__status__ = "Production"
