#!/usr/bin/env python
__all__ = [
    "test_blast",
    "test_blast_xml",
    "test_cigar",
    "test_clustal",
    "test_dialign",
    "test_ebi",
    "test_fasta",
    "test_genbank",
    "test_locuslink",
    "test_ncbi_taxonomy",
    "test_nexus",
    "test_phylip",
    "test_record",
    "test_record_finder",
    "test_tree",
    "test_unigene",
]

__author__ = ""
__copyright__ = "Copyright 2007-2022, The Cogent Project"
__credits__ = [
    "Jeremy Widmann",
    "Catherine Lozuopone",
    "Gavin Huttley",
    "Rob Knight",
    "Sandra Smit",
    "Micah Hamady",
    "Jeremy Widmann",
    "Hua Ying",
    "Greg Caporaso",
    "Zongzhi Liu",
    "Jason Carnes",
    "Peter Maxwell",
    "Marcin Cieslik",
]
__license__ = "BSD-3"
__version__ = "2023.2.12a1"
__maintainer__ = "Gavin Huttley"
__email__ = "Gavin.Huttley@anu.edu.au"
__status__ = "Production"
