#!/usr/bin/env python
__all__ = [
    "test_misc",
    "test_dictarray",
    "test_table",
    "test_transform",
    "test_recode_alignment",
    "test_union_dict",
]

__author__ = ""
__copyright__ = "Copyright 2007-2022, The Cogent Project"
__credits__ = [
    "Jeremy Widmann",
    "Sandra Smit",
    "Gavin Huttley",
    "Rob Knight",
    "Zongzhi Liu",
    "Amanda Birmingham",
    "Greg Caporaso",
]
__license__ = "BSD-3"
__version__ = "2023.2.12a1"
__maintainer__ = "Gavin Huttley"
__email__ = "Gavin.Huttley@anu.edu.au"
__status__ = "Production"
